import unittest
from hypothesis import given, strategies as st

from fastlob import OrderSide
from fastlob.enums import OrderStatus
from fastlob.side import AskSide, BidSide
from fastlob.order import AskOrder, BidOrder, OrderParams
from fastlob.consts import MIN_VALUE, MAX_VALUE

valid_side = st.sampled_from(OrderSide)
valid_price = st.decimals(min_value=MIN_VALUE, max_value=MAX_VALUE-1000, allow_infinity=False, allow_nan=False)
valid_qty = st.decimals(min_value=MIN_VALUE, max_value=MAX_VALUE, allow_infinity=False, allow_nan=False)

class TestSide(unittest.TestCase):
    def setUp(self): pass

    def mkorder(self, params: OrderParams):
        return AskOrder(params) if params.side == OrderSide.ASK else BidOrder(params)

    def test_init(self):
        side = AskSide()

        self.assertTrue(side.empty())
        self.assertEqual(side.side(), OrderSide.ASK)
        self.assertEqual(side.size(), 0)
        self.assertEqual(side.volume(), 0)

        side = BidSide()

        self.assertTrue(side.empty())
        self.assertEqual(side.side(), OrderSide.BID)
        self.assertEqual(side.size(), 0)
        self.assertEqual(side.volume(), 0)

    @given(valid_side, valid_price, valid_qty)
    def test_place_order(self, s, price, qty):
        side = AskSide() if s == OrderSide.ASK else BidSide()

        params = OrderParams(s, price, qty)
        order = self.mkorder(params)

        side.place(order)

        self.assertEqual(side.volume(), order.quantity())
        self.assertEqual(side.size(), 1)
        self.assertFalse(side.empty())
        self.assertEqual(side.best().price(), order.price())
        self.assertEqual(order.status(), OrderStatus.PENDING)

    @given(valid_side, valid_price, valid_qty)
    def test_place_orders(self, s, price, qty):
        side = AskSide() if s == OrderSide.ASK else BidSide()

        for i in range(100):
            params = OrderParams(s, price, qty)
            order = self.mkorder(params)
            side.place(order)

        self.assertEqual(side.volume(), order.quantity()*100)
        self.assertEqual(side.best().volume(), order.quantity()*100)
        self.assertEqual(side.size(), 1)
        self.assertFalse(side.empty())
        self.assertEqual(side.best().price(), order.price())
        self.assertEqual(side.best().valid_orders(), 100)

    def test_ask_best(self):
        side = AskSide()

        params = OrderParams(OrderSide.ASK, 200, 10)
        order = self.mkorder(params)

        side.place(order)

        params = OrderParams(OrderSide.ASK, 100, 10)
        order = self.mkorder(params)
        
        side.place(order)

        self.assertEqual(side.best().price(), 100)
        self.assertEqual(side.size(), 2)

    def test_ask_best_2(self):
        side = AskSide()

        params = OrderParams(OrderSide.ASK, 100, 10)
        order = self.mkorder(params)

        side.place(order)

        params = OrderParams(OrderSide.ASK, 200, 10)
        order = self.mkorder(params)
        
        side.place(order)

        self.assertEqual(side.best().price(), 100)
        self.assertEqual(side.size(), 2)

    def test_bid_best(self):
        side = BidSide()

        params = OrderParams(OrderSide.BID, 200, 10)
        order = self.mkorder(params)

        side.place(order)

        params = OrderParams(OrderSide.BID, 100, 10)
        order = self.mkorder(params)
        
        side.place(order)

        self.assertEqual(side.best().price(), 200)
        self.assertEqual(side.size(), 2)

    def test_bid_best_2(self):
        side = BidSide()

        params = OrderParams(OrderSide.BID, 100, 10)
        order = self.mkorder(params)

        side.place(order)

        params = OrderParams(OrderSide.BID, 200, 10)
        order = self.mkorder(params)
        
        side.place(order)

        self.assertEqual(side.best().price(), 200)
        self.assertEqual(side.size(), 2)

    @given(valid_side, valid_price, valid_qty)
    def test_cancel_order(self, s, price, qty):
        side = AskSide() if s == OrderSide.ASK else BidSide()

        params = OrderParams(s, price, qty)
        order = self.mkorder(params)
        side.place(order)

        side.cancel_order(order)

        self.assertTrue(side.empty())
        self.assertEqual(order.status(), OrderStatus.CANCELED)

    @given(valid_side, valid_price, valid_qty)
    def test_cancel_orders(self, s, price, qty):
        side = AskSide()
        orders = list()

        for i in range(100):
            params = OrderParams(s, price+i, qty)
            order = self.mkorder(params)
            side.place(order)
            orders.append(order)

        self.assertEqual(side.size(), 100)
        self.assertEqual(side.volume(), order.quantity()*100)

        vol = side.volume()

        for i in range(100):
            order = orders[i]
            side.cancel_order(order)

            self.assertEqual(side.size(), 100-i-1)
            self.assertEqual(side.volume(), vol - (order.quantity()*(i+1)))
            self.assertEqual(order.status(), OrderStatus.CANCELED)

        self.assertTrue(side.empty())
        self.assertEqual(side.volume(), 0)



'''The order object manipulated by the lob.'''

import abc
import secrets
from typing import Optional
from decimal import Decimal
from dataclasses import dataclass

from fastlob.enums import OrderSide, OrderType, OrderStatus
from fastlob.consts import ORDERS_ID_SIZE
from .params import OrderParams

@dataclass
class Order(abc.ABC):
    '''Base abstract class for orders in the order-book. Extended by `BidOrder` and `AskOrder`.'''

    _id: str
    _side: OrderSide
    _price: Decimal
    _quantity: Decimal
    _otype: OrderType
    _expiry: Optional[float]
    _status: OrderStatus

    def __init__(self, params: OrderParams):
        self._id       = secrets.token_urlsafe(nbytes=ORDERS_ID_SIZE)
        self._price    = params.price
        self._quantity = params.quantity
        self._otype    = params.otype
        self._expiry   = params.expiry
        self._status   = OrderStatus.CREATED

    def id(self) -> str:
        '''Getter for order identifier.'''
        return self._id

    def side(self) -> OrderSide:
        '''Getter for order side.'''
        return self._side

    def price(self) -> Decimal:
        '''Getter for order price.'''
        return self._price

    def quantity(self) -> Decimal:
        '''Getter for order quantity.'''
        return self._quantity

    def otype(self) -> OrderType:
        '''Getter for order type.'''
        return self._otype

    def expiry(self) -> Optional[float]:
        '''Getter for the expiration date of the order. Only relevant in the case of a GTD order.'''
        return self._expiry

    def status(self) -> OrderStatus:
        '''Getter for order status.'''
        return self._status

    def set_status(self, status: OrderStatus):
        '''Set the order status.'''
        self._status = status

    def fill(self, quantity: Decimal):
        '''Decrease the quantity of the order by some numerical value. If `quantity` is greater than the order qty, 
        we set it to 0.
        '''
        self._quantity -= min(quantity, self._quantity)
        if self.quantity() == 0: self.set_status(OrderStatus.FILLED); return
        self.set_status(OrderStatus.PARTIAL)

    def valid(self) -> bool:
        '''True if order is valid (can be matched).'''
        return self.status() in OrderStatus.valid_states()

    def __eq__(self, other):
        '''Two orders are equal if they're (unique) ids are equal.'''
        return self.id() == other.id()

    def __repr__(self) -> str:
        return f'{self._side.name}Order(id={self.id()}, status={self.status()}, price={self.price()}, ' + \
            f'quantity={self.quantity()}, type={self.otype()})'

@dataclass
class BidOrder(Order):
    '''A bid (buy) order.'''

    def __init__(self, params: OrderParams):
        super().__init__(params)
        self._side = OrderSide.BID

@dataclass
class AskOrder(Order):
    '''An ask (sell) order.'''

    def __init__(self, params: OrderParams):
        super().__init__(params)
        self._side = OrderSide.ASK



'''Order params are used to create orders, they are created by the client.'''

import time
from math import ceil
from decimal import Decimal
from numbers import Number
from typing import Optional

from fastlob.enums import OrderSide, OrderType
from fastlob.utils import todecimal
from fastlob.consts import MIN_VALUE, MAX_VALUE

class OrderParams:
    '''
    This class is used for instantiating orders, it is necessary because we do not want to have the system 
    performing any safety checks, or at least it should have to do as few as possible. 
    Therefore this class is used to force the user to provide valid order attributes.
    '''

    side: OrderSide
    price: Decimal
    quantity: Decimal
    otype: OrderType
    expiry: Optional[int]

    def __init__(self, side: OrderSide, price: Number, quantity: Number, otype: OrderType = OrderType.GTC,
                 expiry: Optional[Number] = None):

        OrderParams.check_args(side, price, quantity, otype, expiry)

        self.side     = side
        self.price    = todecimal(price)
        self.quantity = todecimal(quantity)
        self.otype    = otype
        self.expiry   = int(expiry) if expiry is not None else None

    @staticmethod
    def check_args(side: OrderSide, price: Number, quantity: Number, otype: OrderType, expiry: Optional[Number]):
        '''
        Check for args correctness. 
        This method is very important, since we do not check for this after the object is created.
        If something is wrong it raises the corresponding exception.
        '''

        if not isinstance(side, OrderSide):
            raise TypeError(f'side should of type OrderSide but is {type(side)}')

        if not isinstance(price, Number):
            raise TypeError(f'price should be of type Number but is {type(price)}')

        if not isinstance(quantity, Number):
            raise TypeError(f'quantity should be of type Number but is {type(quantity)}')

        if not isinstance(otype, OrderType):
            raise TypeError(f'ordertype should be of type OrderType but is {type(otype)}')

        if expiry and not isinstance(expiry, Number):
            raise TypeError(f'expiry should be of type Number but is {type(expiry)}')

        if otype == OrderType.GTD:
            if expiry is None: raise ValueError('order is GTD but expiry is None')

            expiry = int(expiry)
            now = ceil(time.time())
            if expiry <= now:
                raise ValueError(f'order expiry ({expiry}) is less than current timestamp ({now}), or too close')

        price_decimal = todecimal(price)
        quantity_decimal = todecimal(quantity)

        if price_decimal < MIN_VALUE: raise ValueError(f'price ({price}) must be greater than {MIN_VALUE}')

        if quantity_decimal < MIN_VALUE: raise ValueError(f'quantity ({quantity}) must be greater than {MIN_VALUE}')

        if price_decimal > MAX_VALUE: raise ValueError(f'price ({price}) is too large')

        if quantity_decimal > MAX_VALUE: raise ValueError(f'quantity ({quantity}) is too large')

    def unwrap(self) -> tuple[Decimal, Decimal, OrderType, Optional[int]]:
        return self.price, self.quantity, self.otype, self.expiry

    def __repr__(self) -> str:
        return f'OrderParams(side={self.side.name}, price={self.price}, qty={self.quantity}, ' + \
            f'type={self.otype}, expiry={self.expiry})'



'''Utility functions for lob.'''

import logging
from typing import Optional

from fastlob.order import Order
from fastlob.result import ResultBuilder
from fastlob.enums import OrderType

# mostly safety checking

def not_running_error(logger: logging.Logger) -> ResultBuilder:
    '''Build the *not running error*.'''

    result = ResultBuilder.new_error()
    errmsg = 'lob is not running (<ob.start> must be called before it can be used)'
    result.add_message(errmsg); logger.error(errmsg)
    return result

def check_limit_order(order: Order) -> Optional[str]:
    '''Check if limit order can be processed.'''

    match order.otype():

        case OrderType.FOK: # FOK order can not be a limit order by definition
            return 'FOK order is not immediately matchable'

    return None



'''All the project enumerations are here.'''

from enum import Enum

class OrderSide(Enum):
    '''The side of an order/limit, can be BID or ASK.'''

    BID = False
    '''The bid (buy) side.'''
    ASK = True
    '''The ask (sell) side.'''

    @staticmethod
    def invert(side):
        '''Invert the side and return it.'''
        return OrderSide.BID if side == OrderSide.ASK else OrderSide.ASK

class OrderType(Enum):
    '''The type of the order, can be FOK, GTC or GTD.'''

    FOK = 1
    '''A fill or kill (FOK) order is a conditional order requiring the transaction to be executed immediately and to 
    its full amount at a stated price. If any of the conditions are broken, then the order must be automatically 
    canceled (kill) right away.'''
    GTC = 2
    '''A Good-Til-Cancelled (GTC) order is an order to buy or sell a stock that lasts until the order is completed 
    or canceled.
    '''
    GTD = 3
    '''A Good-Til-Day (GTD) order is a type of order that is active until its specified date (UTC seconds timestamp), 
    unless it has already been fulfilled or cancelled.
    '''
    FAKE = 4
    '''Used when running lob with historical data.'''

class OrderStatus(Enum):
    '''The status of an order.'''

    CREATED = 1
    '''Order created but not in a limit queue or executed yet.'''
    PENDING = 2
    '''Order in line in limit to be filled but not modified in any ways yet.'''
    FILLED = 3
    '''Order in limit entirely filled, removed from the limit.'''
    PARTIAL = 4
    '''Order in limit partially filled.'''
    CANCELED = 5
    '''Order canceled, can not be fully or partially filled anymore.'''
    ERROR = 6
    '''Set by the lob if the order can not be accepted.'''

    @staticmethod
    def valid_states() -> set: 
        '''Returns the set of states in which an order is considered valid.'''
        return {OrderStatus.CREATED, OrderStatus.PENDING, OrderStatus.PARTIAL}

class ResultType(Enum):
    '''The type of execution result.'''
    LIMIT  = 0
    '''If the order was placed in a limit.'''
    MARKET = 1
    '''If the order was executed as market.'''
    CANCEL = 2
    '''If the operation was an order cancellation.'''
    ERROR  = 3
    '''If the query could not be processed.'''



import time, logging

from fastlob import Orderbook, OrderParams, OrderSide, OrderType

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO) # set maximum logging level 

    lob = Orderbook(name='ABCD', start=True) # create a lob an start it

    # create an order
    params = OrderParams(
        side=OrderSide.BID,
        price=123.32,
        quantity=3.4,
        otype=OrderType.GTD, 
        expiry=time.time() + 120 # order will expire in two minutes
    )

    result = lob(params); assert result.success() # place order

    status, qty_left = lob.get_status(result.orderid()) # query status of order
    print(f'Current order status: {status.name}, quantity left: {qty_left}.\n')

    lob.render() # pretty-print the lob 

    lob.stop() # stop background processes



import unittest, logging
from hypothesis import given, strategies as st

from fastlob import Orderbook, OrderParams, OrderSide, OrderType, OrderStatus, todecimal, ResultType
from fastlob.consts import MIN_VALUE, MAX_VALUE

valid_side = st.sampled_from(OrderSide)
valid_price = st.decimals(min_value=MIN_VALUE, max_value=MAX_VALUE, allow_nan=False, allow_infinity=False)
valid_qty = st.decimals(min_value=MIN_VALUE, max_value=MAX_VALUE, allow_nan=False, allow_infinity=False)

class TestOrdersGTC(unittest.TestCase):
    def setUp(self): 
        logging.basicConfig(level=logging.ERROR)
        
    @given(valid_side, valid_price, valid_qty)
    def test_place_one(self, side, price, qty):
        self.lob = Orderbook('TestGTC')
        self.lob.start()

        self.assertEqual(self.lob.n_prices(), 0)

        params = OrderParams(side, price, qty, OrderType.GTC, expiry=None)
        r = self.lob(params)

        self.assertTrue(r.success())

        self.assertEqual(self.lob.n_prices(), 1)

        self.assertEqual(r.kind(), ResultType.LIMIT)

        if side == OrderSide.ASK:
            self.assertEqual(self.lob.n_asks(), 1)
            self.assertTupleEqual(self.lob.best_ask(), (params.price, params.quantity, 1))
        else:
            self.assertEqual(self.lob.n_bids(), 1)
            self.assertTupleEqual(self.lob.best_bid(), (params.price, params.quantity, 1))

        s, q = self.lob.get_status(r.orderid())

        self.assertEqual(s, OrderStatus.PENDING)
        self.assertEqual(q, params.quantity)

        self.lob.stop()

    def test_place_many(self):
        self.lob = Orderbook('TestGTC')
        self.lob.start()

        self.assertEqual(self.lob.n_prices(), 0)

        params = list()

        N = 40_000

        for i in range(N):
            p = OrderParams(OrderSide.ASK, 50_000+i, 10, OrderType.GTC, expiry=None)
            params.append(p)

            p = OrderParams(OrderSide.BID, 50_000-1-i, 10, OrderType.GTC, expiry=None)
            params.append(p)

        results = self.lob(params)

        self.assertTrue(all([r.success() for r in results]))

        self.assertEqual(self.lob.n_asks(), N)
        self.assertEqual(self.lob.n_prices(), 2*N)

        for r in results:
            self.assertEqual(r.kind(), ResultType.LIMIT)
            s, q = self.lob.get_status(r.orderid())
            self.assertEqual(s, OrderStatus.PENDING)
            self.assertEqual(q, 10)

        self.lob.stop()

    @given(valid_side, valid_price, valid_qty)
    def test_fill_one(self, side, price, qty):
        self.lob = Orderbook('TestGTC')
        self.lob.start()

        self.assertEqual(self.lob.n_prices(), 0)

        params = OrderParams(side, price, qty, OrderType.GTC, expiry=None)
        r = self.lob(params)

        self.assertEqual(r.kind(), ResultType.LIMIT)

        self.assertTrue(r.success())

        self.assertEqual(self.lob.n_prices(), 1)

        matching_order = OrderParams(OrderSide.invert(side), price, qty, OrderType.GTC)

        mr = self.lob(matching_order)

        self.assertDictEqual(mr.execprices(), {params.price: params.quantity})

        self.assertEqual(mr.kind(), ResultType.MARKET)

        self.assertTrue(mr.success())
        self.assertEqual(mr.n_orders_matched(), 1)

        s1, q1 = self.lob.get_status(r.orderid())
        s2, q2 = self.lob.get_status(mr.orderid())

        self.assertEqual(s1, OrderStatus.FILLED)
        self.assertEqual(s2, OrderStatus.FILLED)

        self.assertEqual(q1, 0)
        self.assertEqual(q2, 0)

        self.lob.stop()

    def test_fill_many(self):
        self.lob = Orderbook('TestGTC')
        self.lob.start()

        self.assertEqual(self.lob.n_prices(), 0)

        params = list()

        N = 10

        for i in range(N):
            p = OrderParams(OrderSide.ASK, 50_000+i, 10, OrderType.GTC, expiry=None)
            params.append(p)

        for i in range(N):
            p = OrderParams(OrderSide.BID, 50_000-1-i, 10, OrderType.GTC, expiry=None)
            params.append(p)

        results = self.lob(params)

        self.assertTrue(all([r.success() for r in results]))

        matching_ask = OrderParams(OrderSide.BID, 90_000, 400_000, OrderType.GTC)
        mr1 = self.lob(matching_ask)

        self.assertTrue(mr1.success())
        self.assertEqual(self.lob.n_asks(), 0)

        ep = mr1.execprices()

        for p in params[:len(params)//2]:
            self.assertTrue(p.price in ep.keys())
            self.assertEqual(ep[p.price], todecimal(10))

        matching_bid = OrderParams(OrderSide.ASK, 1, 400_000, OrderType.GTC)
        mr2 = self.lob(matching_bid)

        ep = mr2.execprices()

        for p in params[len(params)//2:]:
            self.assertTrue(p.price in ep.keys())
            self.assertEqual(ep[p.price], todecimal(10))

        self.assertTrue(mr2.success())
        self.assertEqual(self.lob.n_bids(), 0)

        self.assertEqual(mr1.kind(), ResultType.MARKET)
        self.assertEqual(mr2.kind(), ResultType.MARKET)

        for r in results + [mr1, mr2]:
            s, _ = self.lob.get_status(r.orderid())
            self.assertEqual(s, OrderStatus.FILLED)

        self.lob.stop()

    @given(valid_side, valid_price)
    def test_partially_fill_one(self, side, price):
        self.lob = Orderbook('TestGTC')
        self.lob.start()

        self.assertEqual(self.lob.n_prices(), 0)

        params = OrderParams(side, price, 10, OrderType.GTC, expiry=None)
        r = self.lob(params)

        self.assertTrue(r.success())

        matching_order = OrderParams(OrderSide.invert(side), price, 5)

        mr = self.lob(matching_order)

        self.assertEqual(mr.kind(), ResultType.MARKET)

        self.assertDictEqual(mr.execprices(), {params.price: 5})

        self.assertTrue(mr.success())

        s1, q = self.lob.get_status(r.orderid())
        s2, _ = self.lob.get_status(mr.orderid())

        self.assertEqual(s1, OrderStatus.PARTIAL)
        self.assertEqual(q, 5)
        self.assertEqual(s2, OrderStatus.FILLED)

        self.lob.stop()

    @given(valid_side, valid_price)
    def test_fill_limit(self, side, price):
        self.lob = Orderbook('TestGTC')
        self.lob.start()

        self.assertEqual(self.lob.n_prices(), 0)

        params = [OrderParams(side, price, 10, OrderType.GTC, expiry=None)] * 10
        results = self.lob(params)

        self.assertTrue(all([r.success() for r in results]))

        matching_order = OrderParams(OrderSide.invert(side), price, 100)

        mr = self.lob(matching_order)

        self.assertEqual(mr.kind(), ResultType.MARKET)

        self.assertTrue(mr.success())

        self.assertDictEqual(mr.execprices(), {params[0].price: 100})

        self.assertEqual(self.lob.n_prices(), 0)

        self.assertEqual(self.lob.n_prices(), 0)

        for r in results + [mr]:
            s, _ = self.lob.get_status(r.orderid())
            self.assertEqual(s, OrderStatus.FILLED)

        self.lob.stop()

    @given(valid_side, valid_price)
    def test_order_placed_if_not_filled(self, side, price):
        self.lob = Orderbook('TestGTC')
        self.lob.start()

        self.assertEqual(self.lob.n_prices(), 0)

        params = OrderParams(side, price, 10, OrderType.GTC, expiry=None)
        result = self.lob(params)

        self.assertEqual(result.kind(), ResultType.LIMIT)

        self.assertTrue(result.success())

        matching_order = OrderParams(OrderSide.invert(side), price, 15)
        mr = self.lob(matching_order)

        self.assertEqual(mr.kind(), ResultType.MARKET)

        self.assertDictEqual(mr.execprices(), {params.price: 10})

        s, q = self.lob.get_status(mr.orderid())

        self.assertEqual(s, OrderStatus.PENDING)
        self.assertEqual(q, 5)

        s, q = self.lob.get_status(result.orderid())

        self.assertEqual(s, OrderStatus.FILLED)
        self.assertEqual(q, 0)

        self.lob.stop()

    @given(valid_side, valid_price)
    def test_order_placed_if_not_filled2(self, side, price):
        self.lob = Orderbook('TestGTC')
        self.lob.start()

        self.assertEqual(self.lob.n_prices(), 0)

        params = [OrderParams(side, price, 10, OrderType.GTC, expiry=None)]*10
        result = self.lob(params)

        matching_order = OrderParams(OrderSide.invert(side), price, 104.67)
        mr = self.lob(matching_order)

        self.assertEqual(mr.kind(), ResultType.MARKET)

        self.assertDictEqual(mr.execprices(), {params[0].price: 100})

        s, q = self.lob.get_status(mr.orderid())

        self.assertEqual(s, OrderStatus.PENDING)
        self.assertEqual(q, todecimal(4.67))

        for r in result:
            self.assertEqual(mr.kind(), ResultType.MARKET)


            s, q = self.lob.get_status(r.orderid())

            self.assertEqual(s, OrderStatus.FILLED)
            self.assertEqual(q, 0)

        self.lob.stop()


import unittest, logging
from hypothesis import given, strategies as st

from fastlob import Orderbook, OrderParams, OrderSide, OrderType, todecimal
from fastlob.consts import MIN_VALUE, MAX_VALUE

valid_side = st.sampled_from(OrderSide)
valid_price = st.decimals(min_value=MIN_VALUE, max_value=MAX_VALUE, allow_nan=False, allow_infinity=False)
valid_qty = st.decimals(min_value=MIN_VALUE, max_value=MAX_VALUE - MIN_VALUE, allow_nan=False, allow_infinity=False)
valid_qty2 = st.decimals(min_value=MIN_VALUE, max_value=MAX_VALUE // 1000 - MIN_VALUE, allow_nan=False, allow_infinity=False)

class TestOrdersFOK(unittest.TestCase):
    def setUp(self): 
        logging.basicConfig(level=logging.ERROR)

    @given(valid_side, valid_price, valid_qty)
    def test_place_limit(self, side, price, qty):
        # can not place a fok order
        lob = Orderbook('TestOrdersFOK')
        lob.start()
        op = OrderParams(side, price, qty, OrderType.FOK)
        r = lob(op)
        self.assertFalse(r.success())
        lob.stop()
        
    @given(valid_price, valid_qty)
    def test_place_error_qty(self, price, qty):
        # testing that a fok order is not placed if its quantity can not be matched, for one order sitting at one price level

        qty = todecimal(qty)

        lob = Orderbook('TestOrdersFOK')
        lob.start()

        price = 100

        op = OrderParams(OrderSide.ASK, price, qty, OrderType.GTC)
        r = lob(op)

        self.assertTrue(r.success())

        op2 = OrderParams(OrderSide.BID, price, qty + MIN_VALUE, OrderType.FOK)
        r2 = lob(op2)

        self.assertFalse(r2.success())

        op3 = OrderParams(OrderSide.BID, price, quantity=qty, otype=OrderType.FOK)
        r3 = lob(op3)
        self.assertTrue(r3.success())

        lob.stop()
        
    @given(valid_price, valid_qty2)
    def test_place_error_qty2(self, price, qty):
        # testing that a fok order is not placed if its quantity can not be matched, for many orders sitting at one price level

        qty = todecimal(qty)

        lob = Orderbook('TestOrdersFOK')
        lob.start()

        ops = [OrderParams(OrderSide.ASK, price, qty, OrderType.GTC)]*1000
        rs = lob(ops)

        op2 = OrderParams(OrderSide.BID, price, quantity=1000 * qty + MIN_VALUE, otype=OrderType.FOK)
        r2 = lob(op2)

        self.assertFalse(r2.success())

        op3 = OrderParams(OrderSide.BID, price, quantity=1000 * qty, otype=OrderType.FOK)
        r3 = lob(op3)
        self.assertTrue(r3.success())

        lob.stop()

    @given(valid_price, valid_qty2)
    def test_place_error_qty3(self, price, qty):
        # testing that a fok order is not placed if its quantity can not be matched, for many orders sitting at different price levels

        qty = todecimal(qty)

        lob = Orderbook('TestOrdersFOK')
        lob.start()

        price = 100
        
        for i in range(1000):
            op = OrderParams(OrderSide.ASK, price + i, qty, OrderType.GTC)
            r = lob(op)
            self.assertTrue(r.success())

        op2 = OrderParams(OrderSide.BID, price, quantity=qty + MIN_VALUE, otype=OrderType.FOK)
        r2 = lob(op2)
        self.assertFalse(r2.success())

        op3 = OrderParams(OrderSide.BID, price+1000, quantity=1000*qty + MIN_VALUE, otype=OrderType.FOK)
        r3 = lob(op3)
        self.assertFalse(r3.success())

        op4 = OrderParams(OrderSide.BID, price+1000, quantity=1000*qty, otype=OrderType.FOK)
        r4 = lob(op4)
        self.assertTrue(r4.success())

        lob.stop()


'''The result object is returned by the LOB after the client executes an operation.'''

from decimal import Decimal
from typing import Optional
from collections import defaultdict

from fastlob.enums import ResultType

class ResultBuilder:
    '''The object constructed by the lob during execution.'''

    _kind: ResultType
    _orderid: str
    _success: bool
    _messages: list[str]
    _orders_matched: int
    _execprices: Optional[defaultdict[Decimal, Decimal]]

    def __init__(self, kind: ResultType, orderid: str):
        self._kind = kind
        self._orderid = orderid
        self._messages = list()
        self._orders_matched = 0
        self._execprices = defaultdict(Decimal) if kind == ResultType.MARKET else None

    @staticmethod
    def new_limit(orderid: str):
        '''Instantiate a new limit result.'''
        return ResultBuilder(ResultType.LIMIT, orderid)

    @staticmethod
    def new_market(orderid: str):
        '''Instantiate a new market result.'''
        return ResultBuilder(ResultType.MARKET, orderid)

    @staticmethod
    def new_cancel(orderid: str):
        '''Instantiate a new cancel result.'''
        return ResultBuilder(ResultType.CANCEL, orderid)

    @staticmethod
    def new_error():
        '''Instantiate a new error result.'''
        result = ResultBuilder(ResultType.ERROR, None)
        result.set_success(False)
        return result

    def success(self) -> bool:
        '''Getter for success attribute, this attribute should be true if the operation was properly executed.'''
        return self._success

    def set_success(self, success: bool):
        '''Setter for success attribute, this attribute should be true if the operation was properly executed.'''
        self._success = success

    def add_message(self, message: str):
        '''Add an information message destined to the user.'''
        self._messages.append(message)

    def inc_execprices(self, price: Decimal, qty: Decimal):
        '''Increment the number of orders matched at a certain price.'''
        self._execprices[price] += qty

    def inc_orders_matched(self, orders_matched: int):
        '''Increment the total number of orders matched.'''
        self._orders_matched += orders_matched

    def build(self):
        '''Build the ExecutionResult object destined to the client.'''
        return ExecutionResult(self)

class ExecutionResult:
    '''The object returned to the client.'''
    _kind: ResultType
    _orderid: str
    _success: bool
    _messages: list[str]
    _orders_matched: int
    _execprices: Optional[defaultdict[Decimal, Decimal]]

    def __init__(self, result: ResultBuilder):
        self._kind = result._kind
        self._orderid = result._orderid
        self._success = result._success
        self._messages = result._messages
        self._orders_matched = result._orders_matched
        self._execprices = result._execprices

    def kind(self) -> ResultType:
        '''Getter for the result kind, one of LIMIT, CANCEL, MARKET or ERROR.'''
        return self._kind

    def orderid(self) -> str:
        '''Getter for identifier of order executed or canceled.'''
        return self._orderid

    def success(self) -> bool:
        '''Getter for success attribute, true if the operation was executed succesfully.'''
        return self._success

    def messages(self) -> list[str]:
        '''Getter for info messages.'''
        return self._messages.copy()

    def n_orders_matched(self) -> int:
        '''Getter for number of orders matched during execution.'''
        return self._orders_matched

    def execprices(self) -> Optional[defaultdict[Decimal, Decimal]]:
        '''Getter for execprices dict. This dictionary contains the quantity matched at each price level.'''
        return self._execprices.copy()

    def __repr__(self) -> str:
        return f'ExecutionResult(type={self.kind().name}, success={self.success()}, ' + \
            f'orderid={self.orderid()}, messages={self.messages()})'



import unittest
from decimal import Decimal
from hypothesis import given, strategies as st
import time

from fastlob.limit import Limit
from fastlob.enums import OrderSide, OrderStatus, OrderType
from fastlob.order import OrderParams, BidOrder, AskOrder
from fastlob.consts import MIN_VALUE, MAX_VALUE
from fastlob.utils import todecimal

valid_side = st.sampled_from(OrderSide)
valid_price = st.floats(min_value=float(MIN_VALUE), max_value=float(MAX_VALUE), allow_nan=False, allow_infinity=False)
valid_qty = st.floats(min_value=float(MIN_VALUE), max_value=float(MAX_VALUE), allow_nan=False, allow_infinity=False)
valid_otype_noGTD = st.sampled_from([OrderType.FOK, OrderType.GTC])
valid_expiry_noGTD = st.one_of(st.none(), st.floats(min_value=time.time()+5, allow_nan=False, allow_infinity=False))

class TestLimit(unittest.TestCase):
    def setUp(self): pass

    def mkorder(self, params: OrderParams):
        return AskOrder(params) if params.side == OrderSide.ASK else BidOrder(params)

    @given(valid_price)
    def test_init(self, price):
        price = todecimal(price)
        limit = Limit(price)
        self.assertEqual(limit.price(), price)

        self.assertEqual(limit.volume(), 0)
        self.assertEqual(limit.notional(), 0)
        self.assertTrue(limit.empty())
        self.assertTrue(limit.deepempty())
        self.assertEqual(limit.valid_orders(), 0)

    @given(valid_price, valid_side, valid_qty, valid_otype_noGTD, valid_expiry_noGTD)
    def test_enqueue(self, price, side, qty, otype, expiry):
        limit = Limit(price)
        params = OrderParams(side, price, qty, otype, expiry)
        order = self.mkorder(params)

        limit.enqueue(order)

        self.assertEqual(limit.valid_orders(), 1)
        self.assertEqual(limit.next_order(), order)
        self.assertEqual(limit.volume(), order.quantity())

        limit.pop_next_order()

        self.assertTrue(limit.empty())

    @given(valid_price, valid_side, valid_qty, valid_otype_noGTD, valid_expiry_noGTD)
    def test_cancel(self, price, side, qty, otype, expiry):
        limit = Limit(price)
        params = OrderParams(side, price, qty, otype, expiry)
        order = self.mkorder(params)
        limit.enqueue(order)
        limit.cancel_order(order)

        self.assertTrue(limit.empty())
        self.assertEqual(limit.valid_orders(), 0)
        self.assertEqual(order.status(), OrderStatus.CANCELED)

    @given(valid_price, valid_side, valid_qty, valid_otype_noGTD, valid_expiry_noGTD)
    def test_cancel2_A(self, price, side, qty, otype, expiry):
        limit = Limit(price)
        params = OrderParams(side, price, qty, otype, expiry)
        order1 = self.mkorder(params)
        order2 = self.mkorder(params)

        limit.enqueue(order1)
        limit.enqueue(order2)

        self.assertEqual(limit.volume(), order1.quantity() + order2.quantity())

        limit.cancel_order(order1)

        self.assertEqual(limit.valid_orders(), 1)
        self.assertEqual(order2.status(), OrderStatus.PENDING)
        self.assertEqual(order1.status(), OrderStatus.CANCELED)
        self.assertEqual(limit.next_order(), order2)
        self.assertEqual(limit.volume(), order2.quantity())

    @given(valid_price, valid_side, valid_qty, valid_otype_noGTD, valid_expiry_noGTD)
    def test_cancel2_B(self, price, side, qty, otype, expiry):
        limit = Limit(price)
        params = OrderParams(side, price, qty, otype, expiry)
        order1 = self.mkorder(params)
        order2 = self.mkorder(params)

        limit.enqueue(order1)
        limit.enqueue(order2)

        self.assertEqual(limit.volume(), order1.quantity() + order2.quantity())

        limit.cancel_order(order2)

        self.assertEqual(limit.valid_orders(), 1)
        self.assertEqual(order1.status(), OrderStatus.PENDING)
        self.assertEqual(order2.status(), OrderStatus.CANCELED)
        self.assertEqual(limit.next_order(), order1)
        self.assertEqual(limit.volume(), order1.quantity())

    @given(valid_price, valid_side, valid_otype_noGTD, valid_expiry_noGTD)
    def test_fill_next(self, price, side, otype, expiry):
        qty = 10
        limit = Limit(price)
        params = OrderParams(side, price, qty, otype, expiry)
        order = self.mkorder(params)
        limit.enqueue(order)

        limit.fill_next(5)

        self.assertEqual(limit.next_order().quantity(), 5)
        self.assertEqual(limit.next_order().status(), OrderStatus.PARTIAL)

    @given(valid_price, valid_side, valid_qty, valid_otype_noGTD, valid_expiry_noGTD)
    def test_fill_all(self, price, side, qty, otype, expiry):
        limit = Limit(price)
        params = OrderParams(side, price, qty, otype, expiry)

        orders = list()

        for i in range(100):
            order = self.mkorder(params)
            limit.enqueue(order)
            orders.append(order)

        self.assertEqual(limit.valid_orders(), 100)
        self.assertEqual(limit.volume(), order.quantity()*100)

        limit.fill_all()

        self.assertTrue(all([o.quantity() == 0 for o in orders]))
        self.assertTrue(all([o.status() == OrderStatus.FILLED for o in orders]))


'''The engine module is **only** responsible for executing market orders.'''

from decimal import Decimal

from fastlob.side import Side
from fastlob.order import Order
from fastlob.enums import OrderSide
from fastlob.result import ResultBuilder

def execute(order: Order, side: Side) -> ResultBuilder:
    '''Execute a market order at a given side.'''

    result = ResultBuilder.new_market(order.id())

    if fill_whole_limits(side, order, result):
        result.set_success(True); return result

    if fill_whole_orders(side, order, result):
        result.set_success(True); return result

    fill_last_order(side, order, result)

    result.set_success(True); return result

def fill_whole_limits(side: Side, order: Order, result: ResultBuilder) -> bool:
    '''While the order to execute is larger than entire limits, fill them.'''

    while order.quantity() > 0 and not side.empty():
        lim = side.best()

        if oop(order, lim.price()): # if out of price break
            result.add_message(mk_oop_msg(lim.price(), order.quantity()))
            return True

        if order.quantity() < lim.volume(): return False # if can not match whole limits anymore, break

        # update result object
        result.inc_orders_matched(lim.valid_orders())
        result.inc_execprices(lim.price(), lim.volume())

        order.fill(lim.volume()) # partially fill order with limit volume
        side.update_volume(-lim.volume()) # substract limit volume from side volume before filling all orders in limit
        lim.fill_all() # set all orders to filled
        side.pop_limit(lim.price()) # remove limit from side

    return False

def fill_whole_orders(side: Side, order: Order, result: ResultBuilder) -> bool:
    '''While the order to execute is larger than whole orders, fill them.'''

    if side.empty(): return False

    lim = side.best()

    if oop(order, lim.price()):
        result.add_message(mk_oop_msg(lim.price(), order.quantity()))
        return True

    while order.quantity() > 0:
        next_order = lim.next_order()

        if order.quantity() < next_order.quantity(): return False

        result.inc_orders_matched(1)
        result.inc_execprices(next_order.price(), next_order.quantity())

        order.fill(next_order.quantity())
        side.update_volume(-next_order.quantity())
        lim.pop_next_order()

    return False

def fill_last_order(side: Side, order: Order, result: ResultBuilder):
    '''**Partially** fill the last order left with what's left of our order.'''

    if side.empty(): return

    lim = side.best()
    lim_order = lim.next_order()

    if order.valid():
        result.inc_execprices(lim_order.price(), order.quantity())

        lim.fill_next(order.quantity())
        side.update_volume(-order.quantity())

        order.fill(order.quantity())

def oop(order: Order, lim_price: Decimal) -> bool:
    '''True if order is out of price.'''

    match order.side():
        case OrderSide.BID: return order.price() < lim_price
        case OrderSide.ASK: return order.price() > lim_price

def mk_oop_msg(p, q): return f'<matching engine>: order out-of-price at ({p}), quantity left: ({q})'



import unittest, logging, time
from hypothesis import given, strategies as st

from fastlob import Orderbook, OrderStatus, OrderParams, OrderSide, OrderType, todecimal
from fastlob.consts import MIN_VALUE, MAX_VALUE

valid_side = st.sampled_from(OrderSide)
valid_price = st.decimals(min_value=MIN_VALUE, max_value=MAX_VALUE, allow_nan=False, allow_infinity=False)
valid_qty = st.decimals(min_value=MIN_VALUE, max_value=MAX_VALUE - MIN_VALUE, allow_nan=False, allow_infinity=False)
valid_qty2 = st.decimals(min_value=MIN_VALUE, max_value=MAX_VALUE // 1000 - MIN_VALUE, allow_nan=False, allow_infinity=False)

def valid_expiry(x:int = 5): return int(time.time()) + x

class TestOrdersGTD(unittest.TestCase):
    def setUp(self): 
        logging.basicConfig(level=logging.ERROR)

    @given(valid_side, valid_price, valid_qty)
    def test_init(self, side, price, qty):
        with self.assertRaises(ValueError):
            OrderParams(side, price, qty, OrderType.GTD)

        with self.assertRaises(ValueError):
            OrderParams(side, price, qty, OrderType.GTD, expiry=time.time())

        OrderParams(side, price, qty, OrderType.GTD, expiry=valid_expiry())

    @given(valid_side, valid_price, valid_qty)
    def test_place(self, side, price, qty):
        lob = Orderbook('TestOrdersGTD'); lob.start()

        p = OrderParams(side, price, qty, OrderType.GTD, expiry=valid_expiry(10))
        r = lob(p)

        self.assertTrue(r.success())
        self.assertEqual(lob.n_prices(), 1)
        
        s, _ = lob.get_status(r.orderid())
        self.assertEqual(s, OrderStatus.PENDING)

        lob.stop()

    @given(valid_side, valid_price, valid_qty)
    def test_place_then_cancel(self, side, price, qty):
        lob = Orderbook('TestOrdersGTD'); lob.start()

        p = OrderParams(side, price, qty, OrderType.GTD, expiry=valid_expiry(10))
        r = lob(p)

        self.assertTrue(r.success())
        self.assertEqual(lob.n_prices(), 1)
        
        s, _ = lob.get_status(r.orderid())
        self.assertEqual(s, OrderStatus.PENDING)

        cr = lob.cancel(r.orderid())
        self.assertTrue(cr.success())

        s, _ = lob.get_status(r.orderid())
        self.assertEqual(s, OrderStatus.CANCELED)

        lob.stop()

    def test_place_fill(self):
        lob = Orderbook('TestOrdersGTD'); lob.start()

        side = OrderSide.ASK
        price = 100
        qty = 10

        p = OrderParams(side, price, qty, OrderType.GTD, expiry=valid_expiry(2))
        r = lob(p)

        self.assertTrue(r.success())
        self.assertEqual(lob.n_prices(), 1)
        
        s, _ = lob.get_status(r.orderid())
        self.assertEqual(s, OrderStatus.PENDING)

        p2 = OrderParams(OrderSide.invert(side), price, qty)
        r2 = lob(p2)
        self.assertTrue(r2.success())

        s, _ = lob.get_status(r.orderid())
        self.assertEqual(s, OrderStatus.FILLED)

        time.sleep(3.11)

        # check that order is not canceled if it is filled

        s, _ = lob.get_status(r.orderid())
        self.assertEqual(s, OrderStatus.FILLED)

        lob.stop()

    def test_expiration(self):
        self.setUp()
        lob = Orderbook('TestOrdersGTD'); lob.start()

        p = OrderParams(OrderSide.ASK, 100, 10, OrderType.GTD, expiry=valid_expiry(2))
        r = lob(p)

        self.assertTrue(r.success())
        self.assertEqual(lob.n_prices(), 1)
        
        s, _ = lob.get_status(r.orderid())
        self.assertEqual(s, OrderStatus.PENDING)

        time.sleep(3.11)

        s, _ = lob.get_status(r.orderid())
        self.assertEqual(s, OrderStatus.CANCELED)

        lob.stop()

    def test_market_is_not_canceled(self):
        lob = Orderbook('TestOrdersGTD'); lob.start()

        p = OrderParams(OrderSide.ASK, 100, 10, OrderType.GTC)
        r = lob(p)

        self.assertTrue(r.success())
        self.assertEqual(lob.n_prices(), 1)
        s, _ = lob.get_status(r.orderid())
        self.assertEqual(s, OrderStatus.PENDING)

        gtd = OrderParams(OrderSide.BID, 100, 5, OrderType.GTD, expiry=valid_expiry(2))
        rgtd = lob(gtd)
        self.assertTrue(rgtd.success())

        s, _ = lob.get_status(rgtd.orderid())
        self.assertEqual(s, OrderStatus.FILLED)

        time.sleep(3.11)

        s, _ = lob.get_status(rgtd.orderid())
        self.assertEqual(s, OrderStatus.FILLED)

        lob.stop()

    def test_canceled_after_market_partial_fill(self):
        lob = Orderbook('TestOrdersGTD'); lob.start()

        p = OrderParams(OrderSide.ASK, 100, 10, OrderType.GTC)
        r = lob(p)

        self.assertTrue(r.success())
        self.assertEqual(lob.n_prices(), 1)
        s, _ = lob.get_status(r.orderid())
        self.assertEqual(s, OrderStatus.PENDING)

        gtd = OrderParams(OrderSide.BID, 100, 15, OrderType.GTD, expiry=valid_expiry(2))
        rgtd = lob(gtd)
        self.assertTrue(rgtd.success())

        s, _ = lob.get_status(rgtd.orderid())
        self.assertEqual(s, OrderStatus.PENDING)

        time.sleep(3.11)

        s, q = lob.get_status(rgtd.orderid())
        self.assertEqual(s, OrderStatus.CANCELED)
        self.assertEqual(q, 5)

        lob.stop()

    def test_partial_fill_then_cancel(self):
        lob = Orderbook('TestOrdersGTD'); lob.start()

        side = OrderSide.ASK
        price = 100
        qty = 10

        p = OrderParams(side, price, qty, OrderType.GTD, expiry=valid_expiry(2))
        r = lob(p)

        self.assertTrue(r.success())
        self.assertEqual(lob.n_prices(), 1)
        
        s, _ = lob.get_status(r.orderid())
        self.assertEqual(s, OrderStatus.PENDING)

        p2 = OrderParams(OrderSide.invert(side), price, qty // 2)
        r2 = lob(p2)
        self.assertTrue(r2.success())

        s, _ = lob.get_status(r.orderid())
        self.assertEqual(s, OrderStatus.PARTIAL)

        time.sleep(3.11)

        s, _ = lob.get_status(r.orderid())
        self.assertEqual(s, OrderStatus.CANCELED)

        lob.stop()



import unittest, time
from decimal import Decimal
from hypothesis import given
import hypothesis.strategies as st

from fastlob import OrderParams
from fastlob.consts import MIN_VALUE, MAX_VALUE
from fastlob.enums import OrderSide, OrderType
from fastlob.utils import todecimal

valid_side = st.sampled_from(OrderSide)
valid_price = st.floats(min_value=float(MIN_VALUE), max_value=float(MAX_VALUE), allow_nan=False, allow_infinity=False)
valid_qty = st.floats(min_value=float(MIN_VALUE), max_value=float(MAX_VALUE), allow_nan=False, allow_infinity=False)
valid_otype = st.sampled_from(OrderType)
valid_otype_noGTD = st.sampled_from([OrderType.FOK, OrderType.GTC])
valid_expiry = st.one_of(st.floats(min_value=time.time()+5, allow_nan=False, allow_infinity=False))
valid_expiry_noGTD = st.one_of(st.none(), st.floats(min_value=time.time()+5, allow_nan=False, allow_infinity=False))

class TestOrderParams(unittest.TestCase):
    def setUp(self): pass

    @given(valid_side, valid_price, valid_qty, valid_otype_noGTD, valid_expiry_noGTD)
    def test_valid_init_noGTD(self, side, price, qty, otype, expiry):
        params = OrderParams(side, price, qty, otype, expiry)

        self.assertEqual(params.side, side)
        self.assertEqual(params.price, todecimal(price))
        self.assertEqual(params.quantity, todecimal(qty))
        self.assertEqual(params.otype, otype)
        if expiry is None: self.assertEqual(params.expiry, expiry)
        else: self.assertEqual(params.expiry, int(expiry))

    @given(valid_side, valid_price, valid_qty, valid_otype, valid_expiry)
    def test_valid_init(self, side, price, qty, otype, expiry):
        params = OrderParams(side, price, qty, otype, expiry)

        self.assertEqual(params.side, side)
        self.assertEqual(params.price, todecimal(price))
        self.assertEqual(params.quantity, todecimal(qty))
        self.assertEqual(params.otype, otype)
        if expiry is None: self.assertEqual(params.expiry, expiry)
        else: self.assertEqual(params.expiry, int(expiry))

    def test_invalid_side(self):
        with self.assertRaises(TypeError):
            OrderParams(None, 1, 1)

        with self.assertRaises(TypeError):
            OrderParams(42, 1, 1)

    def test_invalid_price(self):
        with self.assertRaises(TypeError):
            OrderParams(OrderSide.ASK, None, 1)

        with self.assertRaises(ValueError):
            OrderParams(OrderSide.ASK, -1, 1)

        with self.assertRaises(ValueError):
            OrderParams(OrderSide.ASK, 0, 1)

    def test_invalid_qty(self):
        with self.assertRaises(TypeError):
            OrderParams(OrderSide.ASK, 1, None)

        with self.assertRaises(ValueError):
            OrderParams(OrderSide.ASK, 1, -1)

        with self.assertRaises(ValueError):
            OrderParams(OrderSide.ASK, 1, 0)

    def test_invalid_otype(self):
        with self.assertRaises(TypeError):
            OrderParams(OrderSide.ASK, 1, 1, None)

        with self.assertRaises(TypeError):
            OrderParams(OrderSide.ASK, 1, 1, 42)

    def test_invalid_expiry(self):
        with self.assertRaises(ValueError):
            OrderParams(OrderSide.ASK, 1, 1, OrderType.GTD, None)

        with self.assertRaises(ValueError):
            OrderParams(OrderSide.ASK, 1, 1, OrderType.GTD, 12)



import unittest, logging
from hypothesis import given, strategies as st

from fastlob import Orderbook, OrderParams, OrderSide, OrderType, OrderStatus
from fastlob.consts import MIN_VALUE, MAX_VALUE

valid_side = st.sampled_from(OrderSide)
valid_price = st.decimals(min_value=MIN_VALUE, max_value=MAX_VALUE, allow_nan=False, allow_infinity=False)
valid_qty = st.decimals(min_value=MIN_VALUE, max_value=MAX_VALUE, allow_nan=False, allow_infinity=False)

class TestCancelGTC(unittest.TestCase):
    def setUp(self): 
        logging.basicConfig(level=logging.ERROR)

    @given(valid_side, valid_price, valid_price)
    def test_cancel_one_limit(self, side, price, qty):
        self.lob = Orderbook('TestCancelGTC')
        self.lob.start()

        p = OrderParams(side, price, qty, OrderType.GTC, expiry=None)

        r = self.lob(p)

        self.assertTrue(r.success())
        self.assertEqual(self.lob.n_prices(), 1)

        cr = self.lob.cancel(r.orderid())

        self.assertTrue(cr.success())
        self.assertEqual(self.lob.n_prices(), 0)

        s, q = self.lob.get_status(cr.orderid())

        self.assertEqual(s, OrderStatus.CANCELED)
        self.assertEqual(q, p.quantity)

        self.lob.stop()

    @given(valid_side, valid_price, valid_price)
    def test_cancel_one_limit_2(self, side, price, qty):
        self.lob = Orderbook('TestCancelGTC')
        self.lob.start()

        p = OrderParams(side, price, qty, OrderType.GTC, expiry=None)

        r = self.lob(p)

        self.assertTrue(r.success())
        self.assertEqual(self.lob.n_prices(), 1)

        p2 = OrderParams(side, price, qty, OrderType.GTC, expiry=None)

        r2 = self.lob(p)

        self.assertTrue(r2.success())
        self.assertEqual(self.lob.n_prices(), 1)

        cr = self.lob.cancel(r.orderid())

        self.assertTrue(cr.success())
        self.assertEqual(self.lob.n_prices(), 1)

        s, q = self.lob.get_status(cr.orderid())

        self.assertEqual(s, OrderStatus.CANCELED)
        self.assertEqual(q, p.quantity)

        cr = self.lob.cancel(r2.orderid())

        self.assertTrue(cr.success())
        self.assertEqual(self.lob.n_prices(), 0)

        s, q = self.lob.get_status(cr.orderid())

        self.assertEqual(s, OrderStatus.CANCELED)
        self.assertEqual(q, p.quantity)

        self.lob.stop()

    def test_cancel_exec(self):
        self.lob = Orderbook('TestCancelGTC')
        self.lob.start()

        p = OrderParams(OrderSide.ASK, 100, 10, OrderType.GTC, expiry=None)
        p2 = OrderParams(OrderSide.ASK, 100, 10, OrderType.GTC, expiry=None)

        r, r2 = self.lob([p, p2])

        self.assertTrue(r.success())
        self.assertTrue(r2.success())
        self.assertEqual(self.lob.n_prices(), 1)

        cr = self.lob.cancel(r.orderid())

        self.assertTrue(cr.success())

        p3 = OrderParams(OrderSide.BID, 100, 15, OrderType.GTC, expiry=None)

        r3 = self.lob(p3)

        self.assertTrue(r3.success())

        self.assertEqual(self.lob.n_bids(), 1)

        self.assertEqual(self.lob.get_status(r3.orderid())[1], 5)

        self.lob.stop()

    def test_cancel_after_partial_fill(self):
        lob = Orderbook('TestCancelGTC')
        lob.start()

        p = OrderParams(OrderSide.ASK, 100, 10, OrderType.GTC, expiry=None)

        r = lob(p)

        self.assertTrue(r.success())
        self.assertEqual(lob.n_asks(), 1)

        partial = OrderParams(OrderSide.BID, 100, 5)

        r2 = lob(partial)

        self.assertTrue(r2.success())
        self.assertEqual(lob.n_asks(), 1)

        s, q = lob.get_status(r.orderid())

        self.assertEqual(s, OrderStatus.PARTIAL)
        self.assertEqual(q, 5)

        cr = lob.cancel(r.orderid())

        self.assertTrue(cr.success())
        self.assertEqual(lob.n_prices(), 0)

        lob.stop()

    def test_cancel_after_complete_fill(self):
        lob = Orderbook('TestCancelGTC')
        lob.start()

        p = OrderParams(OrderSide.ASK, 100, 10, OrderType.GTC, expiry=None)

        r = lob(p)

        self.assertTrue(r.success())
        self.assertEqual(lob.n_asks(), 1)

        full = OrderParams(OrderSide.BID, 100, 10)

        r2 = lob(full)

        self.assertTrue(r2.success())
        self.assertEqual(lob.n_asks(), 0)

        s, q = lob.get_status(r.orderid())

        self.assertEqual(s, OrderStatus.FILLED)
        self.assertEqual(q, 0)

        cr = lob.cancel(r.orderid())

        self.assertFalse(cr.success())

        lob.stop()

    def test_cancel_after_cancel(self):
        lob = Orderbook('TestCancelGTC')
        lob.start()

        p = OrderParams(OrderSide.ASK, 100, 10, OrderType.GTC, expiry=None)

        r = lob(p)

        self.assertTrue(r.success())
        self.assertEqual(lob.n_asks(), 1)

        cr = lob.cancel(r.orderid())

        self.assertTrue(cr.success())

        cr = lob.cancel(r.orderid())

        self.assertFalse(cr.success())

        lob.stop()


import unittest
from decimal import Decimal
from hypothesis import given
import hypothesis.strategies as st
import time

from fastlob import OrderParams
from fastlob.consts import MAX_VALUE, MIN_VALUE
from fastlob.enums import OrderSide, OrderType, OrderStatus
from fastlob.utils import todecimal
from fastlob.order import Order, BidOrder, AskOrder

valid_side = st.sampled_from(OrderSide)
valid_price = st.floats(min_value=float(MIN_VALUE), max_value=float(MAX_VALUE), allow_nan=False, allow_infinity=False)
valid_qty = st.floats(min_value=float(MIN_VALUE), max_value=float(MAX_VALUE), allow_nan=False, allow_infinity=False)
valid_otype_noGTD = st.sampled_from([OrderType.FOK, OrderType.GTC])
valid_expiry_noGTD = st.one_of(st.none(), st.floats(min_value=time.time()+5, allow_nan=False, allow_infinity=False))

class TestOrder(unittest.TestCase):
    def setUp(self): pass

    def mkorder(self, params: OrderParams):
        return AskOrder(params) if params.side == OrderSide.ASK else BidOrder(params)

    @given(valid_side, valid_price, valid_qty, valid_otype_noGTD, valid_expiry_noGTD)
    def test_init(self, side, price, qty, otype, expiry):
        params = OrderParams(side, price, qty, otype, expiry) 

        order = self.mkorder(params)

        self.assertEqual(order.side(), params.side)
        self.assertEqual(order.price(), params.price)
        self.assertEqual(order.quantity(), params.quantity)
        self.assertEqual(order.otype(), params.otype)
        self.assertEqual(order.expiry(), params.expiry)
        self.assertEqual(order.status(), OrderStatus.CREATED)

    @given(valid_side, valid_price, valid_qty, valid_otype_noGTD, valid_expiry_noGTD)
    def test_eq(self, side, price, qty, otype, expiry):
        params = OrderParams(side, price, qty, otype, expiry) 

        order1 = self.mkorder(params)
        order2 = self.mkorder(params)

        self.assertFalse(order1 == order2)

    @given(valid_side, valid_price, valid_qty, valid_otype_noGTD, valid_expiry_noGTD, valid_qty)
    def test_fill(self, side, price, qty, otype, expiry, tofill):
        params = OrderParams(side, price, qty, otype, expiry) 

        order = self.mkorder(params)
        order.fill(order.quantity() + 1)
        self.assertEqual(order.quantity(), 0)
        self.assertEqual(order.status(), OrderStatus.FILLED)

        tofill = todecimal(tofill)

        order = self.mkorder(params)
        qty = order.quantity()
        order.fill(tofill)
        if tofill >= qty: 
            self.assertEqual(0, order.quantity())
            self.assertEqual(order.status(), OrderStatus.FILLED)
        else: 
            self.assertEqual(qty - tofill, order.quantity())
            self.assertEqual(order.status(), OrderStatus.PARTIAL)


'''Utility functions for side.'''

from numbers import Number

def check_update_pair(pair) -> None:
    '''Raise an exception if the pair provided can not be processed as update.'''

    if not isinstance(pair, tuple) or len(pair) != 2:
        raise ValueError('must be pairs of (price, volume)')

    price, volume = pair

    if not isinstance(price, Number) or not isinstance(volume, Number):
        raise ValueError('(price, volume) must be both instances of Number')

    if price <= 0: raise ValueError(f'price must be strictly positive but is {price}')

def check_snapshot_pair(pair):
    '''Raise an exception if the pair provided can not be processed as snapshot.'''

    check_update_pair(pair)

    _, volume = pair
    if volume <= 0: raise ValueError(f'volume must be strictly positive but is {volume}')



import unittest, logging
from hypothesis import given, strategies as st

from fastlob import Orderbook

valid_name = st.text(max_size=1000)
valid_n_snapshot = st.integers(min_value=1, max_value=100)

class TestSide(unittest.TestCase):
    def setUp(self):
        logging.basicConfig(level=logging.FATAL)

    @given(valid_name)
    def test_init(self, name):
        lob = Orderbook(name)

        self.assertEqual(lob._name, name)

        self.assertEqual(lob.best_ask(), None)
        self.assertEqual(lob.best_bid(), None)
        self.assertEqual(lob.midprice(), None)
        self.assertEqual(lob.spread(), None)

    @given(valid_n_snapshot)
    def test_from_snapshot(self, N):

        range_bids = list(range(1, N+1))
        range_asks = list(range(N+2, 2*N+2))

        snapshot = {
            'bids': [(a, b) for a, b in zip(range_bids, range_bids)],
            'asks': [(a, b) for a, b in zip(range_asks, range_asks)],
        }

        lob = Orderbook.from_snapshot(snapshot)

        self.assertEqual(lob.n_prices(), 2*N)

        for a, b in zip(lob.best_bids(lob.n_bids()), list(reversed(snapshot['bids']))):
            self.assertTupleEqual(a[:2], b)

        for a, b in zip(lob.best_asks(lob.n_asks()), snapshot['asks']):
            self.assertTupleEqual(a[:2], b)

    def test_load_updates_and_step(self):
        lob = Orderbook()
        updates = [
            {"bids": [(100, 10), (99, 0)], "asks": [(101, 5), (102, 0)]},
            {"bids": [(101, 15)], "asks": [(100, 4)]},
        ]

        lob.load_updates(updates)

        # First update
        lob.step()
        self.assertEqual(lob.n_bids(), 1)
        self.assertEqual(lob.n_asks(), 1)
        self.assertEqual(lob.best_bid()[:2], (100, 10))
        self.assertEqual(lob.best_ask()[:2], (101, 5))

        # Second update
        lob.step()
        self.assertEqual(lob.n_bids(), 2)
        self.assertEqual(lob.n_asks(), 2)
        self.assertEqual(lob.best_bid()[:2], (101, 15))
        self.assertEqual(lob.best_ask()[:2], (100, 4))

    def test_step_without_load(self):
        lob = Orderbook()
        lob.step()  # Should be a no-op with no crash
        self.assertIsNone(lob.best_bid())
        self.assertIsNone(lob.best_ask())

    def test_step_after_updates_exhausted(self):
        lob = Orderbook()
        updates = [{"bids": [(99, 1)], "asks": [(101, 2)]}]
        lob.load_updates(updates)

        lob.step()  # Applies update
        self.assertEqual(lob.n_bids(), 1)
        self.assertEqual(lob.n_asks(), 1)

        lob.step()  # No-op, should not throw
        self.assertEqual(lob.n_bids(), 1)
        self.assertEqual(lob.n_asks(), 1)


'''A limit is a collection of limit orders sitting at a certain price.'''

from decimal import Decimal
from collections import deque

from fastlob.order import Order
from fastlob.enums import OrderStatus
from fastlob.utils import zero

class Limit:
    '''A limit is a collection of limit orders sitting at a certain price.'''

    _price: Decimal
    _volume: Decimal
    _valid_orders: int
    _orderqueue: deque[Order]
    _fakeorder: Order

    def __init__(self, price: Decimal):
        '''
        Args:
            price (num): The price at which the limit will sit.
        '''
        self._price        = price
        self._volume       = zero()
        self._valid_orders = 0
        self._orderqueue   = deque()
        self._fakeorder    = None

    def price(self) -> Decimal:
        '''Getter for limit price.'''
        return self._price

    def volume(self) -> Decimal:
        '''Getter for limit volume (sum of orders quantity).'''
        return self._volume

    def notional(self) -> Decimal:
        '''Notional = limit price * limit volume.'''
        return self.price() * self.volume()

    def valid_orders(self) -> int:
        '''Getter for limit size (number of orders).'''
        return self._valid_orders

    def empty(self) -> bool:
        '''Check if limit contains zero **valid** orders, not if the limit queue is empty.'''
        return self.valid_orders() == 0

    def deepempty(self):
        '''Check if limit contains zero orders.'''
        return len(self._orderqueue) == 0

    def next_order(self) -> Order:
        '''Returns the next order to be matched by an incoming market order.'''
        self._prune_canceled()
        return self._orderqueue[0]

    def enqueue(self, order: Order):
        '''Add (enqueue) an order to the limit order queue.'''
        self._orderqueue.append(order)
        order.set_status(OrderStatus.PENDING)
        self._volume += order.quantity()
        self._valid_orders += 1

    def fill_next(self, quantity: Decimal):
        '''**Partially** fill the next order in the queue. Filling it entirely would lead to problems, to only use in 
        last stage of order execution (`engine._partial_fill_order`).
        '''
        order = self.next_order()
        order.fill(quantity)
        self._volume -= quantity

    def fill_all(self):
        '''Fill all orders in limit.'''
        while self.valid_orders() > 0:
            order = self.next_order()
            order.fill(order.quantity())
            self.pop_next_order()

    def pop_next_order(self) -> None:
        '''Pop from the queue the next order to be executed. Does not return it, only removes it.'''
        self._prune_canceled()
        order = self._orderqueue.popleft()
        self._valid_orders -= 1
        self._volume -= order.quantity()

    def cancel_order(self, order: Order) -> None:
        '''Cancel an order.'''
        self._volume -= order.quantity()
        self._valid_orders -= 1
        order.set_status(OrderStatus.CANCELED)

    def _prune_canceled(self) -> None:
        '''Pop the next order while it is a canceled one.'''
        while not self.deepempty() and self._orderqueue[0].status() == OrderStatus.CANCELED:
            self._orderqueue.popleft()

    def view(self) -> str:
        '''Returns a pretty-print view of the limit.'''
        return f'{self.price()} | {self.valid_orders():03d} | {self.volume():0>8f} | {self.notional()}'

    def __repr__(self) -> str:
        return f'Limit(price={self.price()}, n_orders={self.valid_orders()}, notional={self.notional()})'

    #### RELATED TO FAKE ORDERS

    def fakeorder_exists(self) -> bool:
        '''True if limit contains a fake order.'''
        return self._fakeorder is not None

    def set_fakeorder(self, order: Order) -> None:
        '''Set `order` as the new limit fake order. This method first deletes the previous order.'''
        self.delete_fakeorder()
        self._fakeorder = order
        self.enqueue(self._fakeorder)

    def delete_fakeorder(self) -> None:
        '''Deletes the current fake order, if no fake order is set it does nothing.'''
        if self._fakeorder is None: return
        self.cancel_order(self._fakeorder)
        self._fakeorder = None



'''Various constants and parameters used in the project.'''

import os
from decimal import Decimal

DECIMAL_PRECISION_ENV_VARNAME: str = 'PYLOB_DECIMAL_PRECISION'

DEFAULT_DECIMAL_PRECISION: int = 2

def _get_precision() -> int:
    precision = os.environ.get(DECIMAL_PRECISION_ENV_VARNAME)
    return int(precision) if precision else DEFAULT_DECIMAL_PRECISION

DECIMAL_PRECISION: int = _get_precision()

MIN_VALUE = UNIT = Decimal('0.' + ('0' * (DECIMAL_PRECISION - 1)) + '1')

MAX_VALUE = Decimal(int(10e10))

ORDERS_ID_SIZE = 8

DEFAULT_LIMITS_VIEW = 10



'''Main module containing the Orderbook class.'''

import io
import time
import logging
import threading
from decimal import Decimal
from typing import Optional, Iterable
from sortedcollections import SortedDict
from termcolor import colored

from fastlob import engine
from fastlob.side import AskSide, BidSide
from fastlob.order import OrderParams, Order, AskOrder, BidOrder
from fastlob.enums import OrderSide, OrderStatus, OrderType
from fastlob.result import ResultBuilder, ExecutionResult
from fastlob.utils import time_asint
from fastlob.consts import DEFAULT_LIMITS_VIEW

from .utils import not_running_error, check_limit_order

class Orderbook:
    '''
    The `Orderbook` is a collection of bid and ask limits. 
    It is reponsible for:
    - Calling `engine` when order is market.
    - Placing order when order is limit.
    - All the safety checking.
    '''

    _name: str
    _askside: AskSide
    _bidside: BidSide
    _orders: dict[str, Order]
    _expirymap: SortedDict
    _start_time: int
    _alive: bool
    _logger: logging.Logger
    _updates: Iterable[dict]

    def __init__(self, name: Optional[str] = 'LOB-1', start: Optional[bool] = False):
        '''
        Args:
            name (Optional[str]): Defaults to "LOB".
        '''
        self._name       = name
        self._askside    = AskSide()
        self._bidside    = BidSide()
        self._orders     = dict()
        self._expirymap  = SortedDict()
        self._start_time = None
        self._alive      = False
        self._updates    = None

        self._logger = logging.getLogger(f'[{name}]')
        self._logger.info('lob initialized, ready to be started using <ob.start>')

        if start: self.start()

    #### MAIN FUNCS

    def start(self):
        '''Start the lob.'''

        def clean_expired_orders():
            while self._alive:
                self._cancel_expired_orders()
                time.sleep(0.1) # what value to set here ? maybe it should depend on the size of the book

        self._alive = True
        self._start_time = time_asint()
        self._logger.info('starting background GTD orders manager..')
        threading.Thread(target=clean_expired_orders).start()
        self._logger.info('lob started properly, ready to receive orders')

    def stop(self):
        '''Stop the lob.'''

        if not self._alive:
            self._logger.error('lob is not running')
            return

        self._alive = False
        self._start_time = None
        self._logger.info('lob stopped properly')

    def reset(self) -> None:
        '''Reset the lob.'''

        if self._alive:
            self._logger.error('lob must be stopped (using <ob.stop>) before reset can be called')
            return

        self.__init__(self._name)

    def __call__(self, order_params: OrderParams | Iterable[OrderParams]) -> ExecutionResult | list[ExecutionResult]:
        '''Process one or many orders: equivalent to calling `process_one` or `process_many`.'''

        if not isinstance(order_params, Iterable):
            return self.process_one(order_params)

        return self.process_many(order_params)

    def process_many(self, orders_params: Iterable[OrderParams]) -> list[ExecutionResult]:
        '''Process many orders at once.

        Args:
            orders_params (Iterable[OrderParams]): Orders to create and process.
        '''
        if not self._alive:
            return [not_running_error(self._logger).build() for _ in orders_params]

        return [self.process_one(params) for params in orders_params]

    def process_one(self, order_params: OrderParams) -> ExecutionResult:
        '''Creates and processes the order corresponding to the corresponding order params.'''

        if not self._alive:
            return not_running_error(self._logger).build()

        if not isinstance(order_params, OrderParams):
            result = ResultBuilder.new_error()
            errmsg = 'order_params is not an instance of fastlob.OrderParams'
            result.add_message(errmsg); self._logger.error(errmsg)
            return result.build()

        #                                         (params const already checks that expiry is set)
        if order_params.otype == OrderType.GTD and order_params.expiry <= (t := time_asint()):
            result = ResultBuilder.new_error()
            errmsg = f'GTD order must expire in the future (but {order_params.expiry} <= {t})'
            result.add_message(errmsg); self._logger.error(errmsg)
            return result.build()

        self._logger.info('processing order params')

        match order_params.side:
            case OrderSide.ASK:
                order = AskOrder(order_params)
                result = self._process_ask_order(order)

            case OrderSide.BID:
                order = BidOrder(order_params)
                result = self._process_bid_order(order)

        if result.success():
            self._logger.info('order %s was processed successfully', order.id())
            self._save_order(order, result)

        else: self._logger.warning('order was not successfully processed')

        if order.status() == OrderStatus.PARTIAL:
            msg = f'order {order.id()} partially filled by engine, {order.quantity()} placed at {order.price()}'
            self._logger.info(msg)
            result.add_message(msg)

        return result.build()

    def cancel(self, orderid: str) -> ExecutionResult:
        '''Cancel an order sitting in the lob given its id.'''

        if not self._alive:
            return not_running_error(self._logger).build()

        self._logger.info('attempting to cancel order with id %s', orderid)

        result = ResultBuilder.new_cancel(orderid)

        try: order = self._orders[orderid]
        except KeyError:
            result.set_success(False)
            errmsg = f'order {orderid} not found in lob'
            result.add_message(errmsg)
            self._logger.warning(errmsg)
            return result.build()

        if not order.valid():
            result.set_success(False)
            errmsg = f'order {orderid} can not be canceled (status={order.status()})'
            result.add_message(errmsg)
            self._logger.warning(errmsg)
            return result.build()

        self._logger.info('order %s can be canceled', orderid)

        match order.side():
            case OrderSide.BID:
                with self._bidside.lock():
                    self._logger.info('cancelling bid order %s', orderid)
                    self._bidside.cancel_order(order)

            case OrderSide.ASK:
                with self._askside.lock():
                    self._logger.info('cancelling ask order %s', orderid)
                    self._askside.cancel_order(order)

        msg = f'order {order.id()} canceled properly'
        result.set_success(True)
        result.add_message(msg)
        self._logger.info(msg)
        return result.build()

    def running_time(self) -> int:
        '''Get time since lob is running.'''

        if not self._alive: return 0
        return time_asint() - self._start_time

    def best_asks(self, n: int) -> list[tuple[Decimal, Decimal, int]]:
        '''Return best `n` asks (price, volume, #orders) triplets. If `n > #asks`, returns `#asks` elements.'''

        if (nasks := self.n_asks()) < n:
            self._logger.warning('asking for %s limits in <ob.best_asks> but lob only contains %s', n, nasks)

        return self._askside.best_limits(n)

    def best_bids(self, n: int) -> list[tuple[Decimal, Decimal, int]]:
        '''Return best `n` bids (price, volume, #orders) triplets. If `n > #bids`, returns `#bids` elements.'''

        if (nbids := self.n_bids()) < n:
            self._logger.warning('asking for %s limits in <ob.best_bids> but lob only contains %s', n, nbids)

        return self._bidside.best_limits(n)

    def best_ask(self) -> Optional[tuple[Decimal, Decimal, int]]:
        '''Get the best ask limit=(price, volume, #orders) in the lob.'''

        if self._askside.empty():
            self._logger.warning('calling <ob.best_ask> but lob does not contain ask limits')
            return None

        lim = self._askside.best()
        return lim.price(), lim.volume(), lim.valid_orders()

    def best_bid(self) -> Optional[tuple[Decimal, Decimal, int]]:
        '''Get the best bid limit=(price, volume, #orders) in the lob.'''

        if self._bidside.empty():
            self._logger.warning('calling <ob.best_bid> but lob does not contain ask limits')
            return None

        lim = self._bidside.best()
        return lim.price(), lim.volume(), lim.valid_orders()

    def n_bids(self) -> int:
        '''Get the number of bids limits.'''

        return self._bidside.size()

    def n_asks(self) -> int:
        '''Get the number of asks limits.'''

        return self._askside.size()

    def n_prices(self) -> int:
        '''Get the total number of limits (price levels).'''

        return self.n_asks() + self.n_bids()

    def bids_volume(self) -> Decimal:
        '''Total volume on the bid side.'''

        return self._bidside.volume()

    def asks_volume(self) -> Decimal:
        '''Total volume on the ask side.'''

        return self._askside.volume()

    def midprice(self) -> Optional[Decimal]:
        '''Get the lob midprice.'''

        if self._askside.empty() or self._bidside.empty():
            self._logger.warning('calling <ob.midprice> but lob does not contain limits on both sides')
            return None

        askprice, bidprice = self.best_ask()[0], self.best_bid()[0]
        return Decimal(0.5) * (askprice + bidprice)

    def spread(self) -> Decimal:
        '''Get the lob spread.'''

        if self._askside.empty() or self._bidside.empty():
            self._logger.warning('calling <ob.spread> but lob does not contain limits on both sides')
            return None

        askprice, bidprice = self.best_ask()[0], self.best_bid()[0]
        return askprice - bidprice

    def get_status(self, orderid: str) -> Optional[tuple[OrderStatus, Decimal]]:
        '''Get the status and the quantity left for a given order or None if order was not accepted by the lob.'''

        try:
            order = self._orders[orderid]
            self._logger.info('order %s found in lob', orderid)
            return order.status(), order.quantity()
        except KeyError:
            self._logger.warning('order %s not found in lob', orderid)
            return None

    def view(self, n : int = DEFAULT_LIMITS_VIEW) -> str:
        '''Get a pretty-printed view of the lob state.'''

        length = 40
        if not self._bidside.empty(): length = len(self._bidside.best().view()) + 2
        elif not self._askside.empty(): length = len(self._askside.best().view()) + 2

        buffer = io.StringIO()
        buffer.write(f"   [ORDER-BOOK {self._name.upper()}]\n\n")
        buffer.write(colored(self._askside.view(n), "red"))
        buffer.write(' ' + '~'*length + '\n')
        buffer.write(colored(self._bidside.view(n), "green"))

        if self._askside.empty() or self._bidside.empty():
            return buffer.getvalue()

        footer =  f'\n - spread = {self.spread()}'
        footer += f' | midprice = {self.midprice()}'
        footer += f'\n - asks volume = {self.asks_volume()}'
        footer += f' | bids volume = {self.bids_volume()}'

        buffer.write(colored(footer, color='blue'))

        return buffer.getvalue()

    def render(self) -> None:
        '''Pretty-print the lob.'''

        print(self.view(), flush=True)

    def __repr__(self) -> str:
        buffer = io.StringIO()
        buffer.write(f'Order-book {self._name}\n')
        buffer.write(f'- started={self._alive}\n')
        buffer.write(f'- running-time={self.running_time()}s\n')
        buffer.write(f'- #prices={self.n_prices()}\n')
        buffer.write(f'- #asks={self.n_asks()}\n')
        buffer.write(f'- #bids={self.n_bids()}')
        return buffer.getvalue()

    #### SNAPSHOT RELATED FUNCS

    @staticmethod
    def from_snapshot(snapshot: dict, name: Optional[str] = 'LOB', start: Optional[bool] = False):
        '''
        Instantiate a new LOB from a given snapshot. A "snapshot" is a dictionary of the following 
        form `{"bids": <list_of_(price, volume)_pairs>, "asks": <list_of_(price, volume)_pairs>}`.

        It does so by placing a "fake" order. These orders are also not added to the history. 
        They are simply added to each price level.

        Returns:
            Orderbook: A new LOB initialized with `snapshot`.
        '''

        if not isinstance(snapshot, dict) or snapshot.keys() != {'bids', 'asks'}:
            raise ValueError('snapshot must be a dictionary containing "bids" and "asks" keys')

        if not isinstance(snapshot['bids'], Iterable) or not isinstance(snapshot['asks'], Iterable):
            raise ValueError('snapshot[bids|asks] must be an iterable of (price, volume) pairs')

        lob = Orderbook(name=name, start=False)

        asks, bids = snapshot['asks'], snapshot['bids']

        lob._askside.apply_snapshot(asks)
        lob._bidside.apply_snapshot(bids)

        if start: lob.start()
        return lob

    #### UPDATES RELATED FUNCS

    def load_updates(self, updates: Iterable[dict]):
        '''Load `updates` so that every time `step` is called, the lob gets updated (using fake orders).'''

        if not isinstance(updates, Iterable):
            raise ValueError(
                'updates must be an Iterable of dicts of the form ' + 
                '{"bids|asks": <iterable_of_(price, volume)_pairs>}')

        self._updates = iter(updates)

    def step(self):
        '''Apply the updates in `next(updates)` to the lob.'''

        if self._updates is None:
            self._logger.warning('calling <ob.step> but nothing was loaded using <ob.load_updates>')
            return

        try: updates = next(self._updates)
        except StopIteration:
            self._logger.warning('calling <ob.step> but iterator is exhausted')
            return

        if not isinstance(updates, dict) or updates.keys() != {'bids', 'asks'}:
            raise ValueError('updates must be a dictionary containing "bids" and "asks" keys')

        bids, asks = updates['bids'], updates['asks']

        # lock all to aply updates
        with self._askside.lock(), self._bidside.lock():

            # apply updates to ask side
            self._askside.apply_updates(asks)

            # apply updates to bid side
            self._bidside.apply_updates(bids)

    #### AUXILIARY FUNCS (where most of the work happens)

    def _process_bid_order(self, order: BidOrder) -> ResultBuilder:
        self._logger.info('processing bid order %s', order.id())

        if self._askside.is_market(order):
            self._logger.info('bid order %s is market', order.id())

            if (error := self._askside.check_market_order(order)) is not None:
                order.set_status(OrderStatus.ERROR)
                result = ResultBuilder.new_market(order.id())
                result.set_success(False)
                result.add_message(error)
                return result

            # execute the order
            with self._askside.lock():
                result = engine.execute(order, self._askside)

            if not result.success():
                self._logger.error('bid market order %s could not be executed by engine', order.id())
                return result

            if order.status() == OrderStatus.PARTIAL:
                with self._bidside.lock():
                    self._bidside.place(order)
                    msg = f'order {order.id()} partially executed, {order.quantity()} was placed as a bid limit order'
                    self._logger.info(msg)
                    result.add_message(msg)

            self._logger.info('executed bid market order %s', order.id())
            return result

        # else: is limit order
        self._logger.info('bid order %s is limit', order.id())

        result = ResultBuilder.new_limit(order.id())

        if (error := check_limit_order(order)) is not None:
            order.set_status(OrderStatus.ERROR)
            result.set_success(False)
            result.add_message(error)
            self._logger.warning(error)
            return result

        # place the order in the side
        with self._bidside.lock(): self._bidside.place(order)

        result.set_success(True)
        self._logger.info('order %s successfully placed', order.id())
        return result

    def _process_ask_order(self, order: AskOrder) -> ResultBuilder:
        self._logger.info('processing ask order %s', order.id())

        if self._bidside.is_market(order):
            self._logger.info('ask order %s is market', order.id())

            if (error := self._bidside.check_market_order(order)) is not None:
                order.set_status(OrderStatus.ERROR)
                result = ResultBuilder.new_market(order.id())
                result.set_success(False)
                result.add_message(error)
                return result

            # execute the order
            with self._bidside.lock():
                result = engine.execute(order, self._bidside)

            if not result.success():
                self._logger.error('ask market order %s could not be executed by engine', order.id())
                return result

            if order.status() == OrderStatus.PARTIAL:
                with self._askside.lock():
                    self._askside.place(order)
                    msg = f'order {order.id()} partially executed, {order.quantity()} was placed as an ask limit order'
                    self._logger.info(msg)
                    result.add_message(msg)

            self._logger.info('executed ask market order %s', order.id())
            return result

        # else is limit order
        self._logger.info('ask order %s is limit', order.id())

        result = ResultBuilder.new_limit(order.id())

        if (error := check_limit_order(order)) is not None:
            order.set_status(OrderStatus.ERROR)
            result.set_success(False)
            result.add_message(error)
            self._logger.warning(error)
            return result

        # place the order in the side
        with self._askside.lock(): self._askside.place(order)

        result.set_success(True)
        self._logger.info('order %s successfully placed', order.id())
        return result

    def _save_order(self, order: Order, result: ResultBuilder):
        self._logger.info('adding order to history')
        self._orders[order.id()] = order

        if order.otype() == OrderType.GTD: # and result._KIND == ResultType.LIMIT:
            # <- TODO: doesnt work in the case where the order is a partially filling market (then placed in limit),
            # but how to not add market orders then ?
            self._logger.info('order is a limit GTD order, adding order to expiry map')
            if order.expiry() not in self._expirymap.keys(): self._expirymap[order.expiry()] = list()
            self._expirymap[order.expiry()].append(order)

    def _cancel_expired_orders(self):
        '''Background expired orders cleaner.'''

        timestamps = self._expirymap.keys()
        if not timestamps: return

        now = time_asint()
        keys_outdated = filter(lambda timestamp: timestamp < now, timestamps)

        for key in keys_outdated:
            expired_orders = self._expirymap[key]

            self._logger.info('GTD orders: cancelling %s with t=%s', len(expired_orders), key)

            for order in expired_orders:
                if not order.valid(): continue

                match order.side():
                    case OrderSide.ASK:
                        with self._askside.lock(): self._askside.cancel_order(order)

                    case OrderSide.BID:
                        with self._bidside.lock(): self._bidside.cancel_order(order)

            del self._expirymap[key]



'''The Side is a collection of limits, whose ordering (by price) depends wether it is a bid or ask side.'''

import io
import abc
import threading
from numbers import Number
from typing import Optional, Iterable
from decimal import Decimal
from collections.abc import Sequence
from sortedcollections import SortedDict

from fastlob.limit import Limit
from fastlob.order import Order, BidOrder, AskOrder, OrderParams
from fastlob.utils import zero, todecimal
from fastlob.enums import OrderSide, OrderType

from .utils import check_snapshot_pair, check_update_pair

class Side(abc.ABC):
    '''The Side is a collection of limits, whose ordering (by price) depends wether it is a bid or ask side.'''

    _side: OrderSide
    _volume: Decimal
    _price2limits: SortedDict[Decimal, Limit]
    _mutex: threading.Lock
    # ^ the role of this mutex is to prevent a limit order being canceled meanwhile we are matching a market order
    # it must be locked by any other class before it can execute or cancel an order in the side

    def __init__(self):
        self._volume = zero()
        self._mutex = threading.Lock()

    def lock(self):
        '''Returns the side mutex lock.'''

        return self._mutex

    def side(self) -> OrderSide:
        '''Get the side of the limit.'''

        return self._side

    def volume(self) -> Decimal:
        '''Getter for side volume, that is the sum of the volume of all limits.'''

        return self._volume

    def update_volume(self, update: Decimal) -> None:
        '''Add `update` to current side volume.'''

        self._volume += update

    def size(self) -> int:
        '''Get number of limits in the side.'''

        return len(self._price2limits)

    def empty(self) -> bool:
        '''Check if side is empty (does not contain any limit).'''

        return self.size() == 0

    def best(self) -> Limit:
        '''Get the best limit of the side.'''

        return self._price2limits.peekitem(0)[1]

    def best_limits(self, n: int) -> list[tuple[Decimal, Decimal, int]]:
        '''Returns a triplet (price, volume, #orders) for the best `n` price levels.'''

        result = list()

        for i, lim in enumerate(self.limits()):
            if i >= n: break
            t = (lim.price(), lim.volume(), lim.valid_orders())
            result.append(t)

        return result

    def limits(self) -> Sequence:
        '''Get all limits (sorted).'''

        return self._price2limits.values()

    def place(self, order: Order) -> None:
        '''Place an order in the side at its corresponding limit.'''

        price = order.price()
        self._new_price_if_not_exists(price)
        self.get_limit(price).enqueue(order)
        self._volume += order.quantity()

    def cancel_order(self, order: Order) -> None:
        '''Cancel an order sitting in the side.'''

        self._volume -= order.quantity()
        lim = self.get_limit(order.price())
        lim.cancel_order(order)
        if lim.empty(): del self._price2limits[lim.price()]

    def get_limit(self, price: Decimal) -> Limit:
        '''Get the limit sitting at a certain price.'''

        return self._price2limits[price]

    def pop_limit(self, price) -> None:
        '''Delete a limit from the side.'''

        self._price2limits.pop(price) # remove limit from side

    def check_market_order(self, order: Order) -> Optional[str]:
        '''Check if a market order is valid.'''

        match order.otype():
            case OrderType.FOK: # check that order quantity can be filled
                if not self.immediately_matched(order):
                    return 'FOK bid order is not immediately matchable'
        return None

    def _price_exists(self, price: Decimal) -> bool:
        '''Check there is a limit at a certain price.'''

        return price in self._price2limits.keys()

    def _new_price(self, price: Decimal) -> None:
        '''Create a new price level in the side.'''

        self._price2limits[price] = Limit(price)

    def _new_price_if_not_exists(self, price: Decimal) -> None:
        '''Create new price level if doesn't exist.'''

        if not self._price_exists(price): self._new_price(price)

    def __repr__(self) -> str:
        if self.empty(): return f'{self.side().name}Side(size={self.size()}, volume={self.volume()})'
        return f'{self.side().name}Side(size={self.size()}, volume={self.volume()}, best={self.best()})'

    @abc.abstractmethod
    def is_market(self, order: Order) -> bool:
        '''Check if an order of the opposite side is market.'''

    @abc.abstractmethod
    def immediately_matched(self, order: Order) -> bool:
        '''Check that a market order (of the opposite side) can be immediately matched. This function is useful
        when checking that a FOK order is valid.'''

    @abc.abstractmethod
    def apply_snapshot(self, snapshot: Iterable[tuple[Number, Number]]):
        '''Initialize side with predefined volume for price levels.'''

    @abc.abstractmethod
    def apply_updates(self, updates: Iterable[tuple[Number, Number]]):
        '''Apply price level updates to side.'''

    @abc.abstractmethod
    def view(self, n : int) -> str:
        '''Get a pretty-printed view of the side.'''

    #### RELATED TO FAKE ORDERS

    def place_fakeorder(self, order: Order):
        '''Place a fake order in the side.'''

        if not self._price_exists(order.price()):
            self._new_price(order.price())

        limit = self.get_limit(order.price())
        prev_limit_volume = limit.volume()

        limit.set_fakeorder(order)
        self.update_volume(limit.volume() - prev_limit_volume)

    def delete_fakeorder(self, price: Decimal):
        '''Delete a fake order at price level `price`.'''

        if not self._price_exists(price): return

        limit = self.get_limit(price)
        if not limit.fakeorder_exists(): return

        limit.delete_fakeorder()
        if limit.volume() == 0.0: self.pop_limit(price)

class BidSide(Side):
    '''The bid side, where **the best price level is the highest**.'''

    def __init__(self):
        super().__init__()
        self._side = OrderSide.BID
        self._price2limits = SortedDict(lambda x: -x)

    def is_market(self, order: AskOrder) -> bool:
        if self.empty(): return False
        if self.best().price() >= order.price(): return True
        return False

    def immediately_matched(self, order: AskOrder) -> bool:
        # we want the limit volume down to the order price to be >= order quantity
        volume = zero()

        lim : Limit
        for lim in self.limits():
            if lim.price() < order.price(): break
            if volume >= order.quantity(): break
            volume += lim.volume()

        if volume < order.quantity(): return False
        return True

    def apply_snapshot(self, bids):
        # apply snapshot (init side) to askside
        for pair in bids:
            check_snapshot_pair(pair)
            price, volume = pair

            params = OrderParams(OrderSide.BID, price, volume, OrderType.FAKE)
            order  = BidOrder(params)
            self.place_fakeorder(order)

    def apply_updates(self, bids):
        # apply updates to bid side
        for pair in bids:
            check_update_pair(pair)
            price, volume = pair

            price  = todecimal(price)
            volume = todecimal(volume)

            if volume == 0:
                self.delete_fakeorder(price)
                continue

            params = OrderParams(OrderSide.BID, price, volume, OrderType.FAKE)
            order  = BidOrder(params)
            self.place_fakeorder(order)

    def view(self, n : int = 10) -> str:
        if self.empty(): return str()

        buffer = io.StringIO()
        count = 0
        for bidlim in self._price2limits.values():
            if count >= n:
                if count < self.size():
                    buffer.write(f"   ...({self.size() - n} more bids)\n")
                break
            buffer.write(f" - {bidlim.view()}\n")
            count += 1

        return buffer.getvalue()

class AskSide(Side):
    '''The bid side, where **the best price level is the lowest**.'''

    def __init__(self):
        super().__init__()
        self._side = OrderSide.ASK
        self._price2limits = SortedDict()

    def is_market(self, order: BidOrder) -> bool:
        if self.empty(): return False
        if self.best().price() <= order.price(): return True
        return False

    def immediately_matched(self, order: BidOrder) -> bool:
        # we want the limit volume down to the order price to be >= order quantity
        volume = zero()
        limits = self.limits()

        lim : Limit
        for lim in limits:
            if lim.price() > order.price(): break
            if volume >= order.quantity(): break
            volume += lim.volume()

        if volume < order.quantity(): return False
        return True

    def apply_snapshot(self, asks):
        # apply snapshot (init side) to askside
        for pair in asks:
            check_snapshot_pair(pair)
            price, volume = pair

            params = OrderParams(OrderSide.ASK, price, volume, OrderType.FAKE)
            order  = AskOrder(params)
            self.place_fakeorder(order)

    def apply_updates(self, asks):
        # apply updates to ask side
        for pair in asks:
            check_update_pair(pair)
            price, volume = pair

            price  = todecimal(price)
            volume = todecimal(volume)

            if volume == 0:
                self.delete_fakeorder(price)
                continue

            params = OrderParams(OrderSide.ASK, price, volume, OrderType.FAKE)
            order  = AskOrder(params)
            self.place_fakeorder(order)

    def view(self, n : int = 10) -> str:
        if self.empty(): return str()

        buffer = io.StringIO()
        if self.size() > n: buffer.write(f"   ...({self.size() - n} more asks)\n")
        count = 0
        l = list()
        for asklim in self._price2limits.values():
            if count >= n: break
            l.append(f" - {asklim.view()}\n")
            count += 1

        buffer.writelines(reversed(l))
        return buffer.getvalue()



'''Global utility functions.'''

import time
from decimal import Decimal
from numbers import Number

from fastlob.consts import DECIMAL_PRECISION

def todecimal(n: Number | str) -> Decimal:
    '''Wrapper around the Decimal constructor to properly round numbers to user defined precision.'''
    if not isinstance(n, Number | str): raise TypeError("invalid type to be converted to decimal")

    dec = Decimal.from_float(n) if isinstance(n, float) else Decimal(n)
    exp = Decimal(f'0.{"0"*DECIMAL_PRECISION}')

    return dec.quantize(exp)

def zero(): return Decimal('0')

def time_asint() -> int:
    '''Return rounded time.time() as int.'''
    return int(time.time())
